﻿using System;
using System.ComponentModel.Composition.Hosting;
using System.ComponentModel.Composition.Primitives;
using System.Linq;
using Microsoft.Practices.Unity;
using System.Reflection;

namespace MachinaAurum.Unity.Mef
{
    public class Container : IUnityContainer
    {
        IUnityContainer Unity;

        UnityAggregatedCatalog AggregatedCatalog;
        CompositionContainer Mef;

        public Container( params Assembly[] assemblies )
        {
            Unity = new UnityContainer();

            var catalogs = assemblies.Select(x => new AssemblyCatalog(x)).ToArray();

            AggregatedCatalog = new UnityAggregatedCatalog(this, catalogs);
            Mef = new CompositionContainer(AggregatedCatalog);

            Unity.AddExtension(new MefUnityContainerExtension(Mef));

            RegisterFuncFactory();
        }

        public Container(params ComposablePartCatalog[] catalogs)
        {
            Unity = new UnityContainer();

            if (catalogs.Any() == false)
            {
                catalogs = AppDomain.CurrentDomain.GetAssemblies().Select(x => new AssemblyCatalog(x)).ToArray();
            }

            UnityAggregatedCatalog aggCatalog = new UnityAggregatedCatalog(this, catalogs);
            Mef = new CompositionContainer(aggCatalog);

            Unity.AddExtension(new MefUnityContainerExtension(Mef));

            RegisterFuncFactory();
        }

        private Container(IUnityContainer unity, CompositionContainer mef)
        {
            Unity = unity;

            ComposablePartCatalog[] catalogs;

            if (mef != null)
            {
                catalogs = (mef.Catalog as UnityAggregatedCatalog).Catalogs.ToArray();
            }
            else
            {
                catalogs = AppDomain.CurrentDomain.GetAssemblies().Select(x => new AssemblyCatalog(x)).ToArray();
            }

            UnityAggregatedCatalog aggCatalog = new UnityAggregatedCatalog(unity, catalogs);
            Mef = new CompositionContainer(aggCatalog);

            unity.AddExtension(new MefUnityContainerExtension(Mef));

            RegisterFuncFactory();
        }

        public void RegisterFuncFactory()
        {
            Unity.RegisterInstance<Func<Type, object>>("Resolve", (type) => this.Resolve(type));
            Unity.RegisterInstance<Action<Type, Type>>("RegisterType", (from, to) => this.RegisterType(from, to ));
            Unity.RegisterInstance<Action<object>>("BuildUp", (x) => this.BuildUp(x.GetType(), x));
        }

        #region IUnityContainer Members

        public IUnityContainer AddExtension(UnityContainerExtension extension)
        {
            Unity.AddExtension(extension);
            return this;
        }

        public object BuildUp(System.Type t, object existing, string name, params ResolverOverride[] resolverOverrides)
        {
            return Unity.BuildUp(t, existing, name, resolverOverrides);
        }

        public object Configure(System.Type configurationInterface)
        {
            return Unity.Configure(configurationInterface);
        }

        public IUnityContainer CreateChildContainer()
        {
            return new Container(Unity.CreateChildContainer(), Mef);
        }

        public IUnityContainer Parent
        {
            get
            {
                return Unity.Parent;
            }
        }

        public IUnityContainer RegisterInstance(System.Type t, string name, object instance, LifetimeManager lifetime)
        {
            Unity.RegisterInstance(t, name, instance, lifetime);
            AggregatedCatalog.AddPart(t, instance);
            return this;
        }

        public IUnityContainer RegisterType(System.Type from, System.Type to, string name, LifetimeManager lifetimeManager, params InjectionMember[] injectionMembers)
        {
            Unity.RegisterType(from, to, name, lifetimeManager, injectionMembers);
            return this;
        }

        public System.Collections.Generic.IEnumerable<ContainerRegistration> Registrations
        {
            get
            {
                return Unity.Registrations;
            }
        }

        public IUnityContainer RemoveAllExtensions()
        {
            Unity.RemoveAllExtensions();
            return this;
        }

        public object Resolve(System.Type t, string name, params ResolverOverride[] resolverOverrides)
        {
            return Unity.Resolve(t, name, resolverOverrides);
        }

        public System.Collections.Generic.IEnumerable<object> ResolveAll(System.Type t, params ResolverOverride[] resolverOverrides)
        {
            return Unity.ResolveAll(t, resolverOverrides);
        }

        public void Teardown(object o)
        {
            Unity.Teardown(o);
        }

        #endregion

        #region IDisposable Members

        public void Dispose()
        {
            Unity.Dispose();
            Mef.Dispose();
        }

        #endregion
    }
}