﻿using System;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using MachinaAurum.Xna.Graphics;
using MachinaAurum.Xna.Graphics.Effects;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace MachinaAurum.Xna.DirectX10
{
    /// <summary>
    /// The Direct X 10 device manager.
    /// </summary>
    [GraphicDevice("DirectX10")]
    public class DirectX10GraphicDevice : IGraphicDevice
    {
        Game Game { get; set; }

        SlimDX.Direct3D10.Device _Device;
        SlimDX.DXGI.SwapChain _SwapChain;

        public SlimDX.Direct3D10.RenderTargetView RenderTargetView { get; private set; }
        public SlimDX.Direct3D10.DepthStencilView DepthStencilView { get; private set; }
        public SlimDX.Direct3D10.Viewport Viewport { get; private set; }

        public SlimDX.Direct3D10.Device Device
        {
            get
            {
                return _Device;
            }
        }

        public SlimDX.DXGI.SwapChain SwapChain
        {
            get
            {
                return _SwapChain;
            }
        }

        #region IGraphicDevice Members

        public IAurumVertexBuffer CreateVertexBuffer(VertexDeclaration vertexDeclaration, int vertexCount)
        {
            return new DirectX10VertexBuffer(this, vertexDeclaration, vertexCount);
        }

        public IAurumIndexBuffer CreateIndexBuffer(int indexSize, int indexCount)
        {
            return new DirectX10IndexBuffer(this, indexSize, indexCount);
        }

        public IAurumEffect CreateEffect(IEffectCode code)
        {
            return new DirectX10Effect(this, code);
        }

        public void Create(Game game)
        {
            Game = game;

            int width = Game.Window.ClientBounds.Width;
            int height = Game.Window.ClientBounds.Height;

            SetUpDevice(width, height);
            SetUpRenderTargetView();
            SetUpDepthStencilBuffer(width, height);
            SetUpRasterizer(width, height);
            SetUpOutputMerger();

            SlimDX.DXGI.Factory factory = SwapChain.GetParent<SlimDX.DXGI.Factory>();
            factory.SetWindowAssociation(Game.Window.Handle, SlimDX.DXGI.WindowAssociationFlags.IgnoreAll);
        }

        private void SetUpRenderTargetView()
        {
            using (var backBuffer = SlimDX.Direct3D10.Texture2D.FromSwapChain<SlimDX.Direct3D10.Texture2D>(SwapChain, 0))
            {
                RenderTargetView = new SlimDX.Direct3D10.RenderTargetView(Device, backBuffer);
            }
        }

        private void SetUpDevice(int width, int height)
        {
            var swapChainDescription = new SlimDX.DXGI.SwapChainDescription()
            {
                IsWindowed = true,
                BufferCount = 1,
                SwapEffect = SlimDX.DXGI.SwapEffect.Discard,
                Usage = SlimDX.DXGI.Usage.RenderTargetOutput,
                ModeDescription = new SlimDX.DXGI.ModeDescription
                (
                    width: width,
                    height: height,
                    refreshRate: new SlimDX.Rational(60, 1),
                    format: SlimDX.DXGI.Format.R8G8B8A8_UNorm
                ),
                OutputHandle = Game.Window.Handle,
                SampleDescription = new SlimDX.DXGI.SampleDescription(count: 1, quality: 0),
            };

            SlimDX.Direct3D10.Device.CreateWithSwapChain
            (
                adapter: null,
                driverType: SlimDX.Direct3D10.DriverType.Hardware,
                flags: SlimDX.Direct3D10.DeviceCreationFlags.Debug,
                swapChainDescription: swapChainDescription,
                device: out _Device,
                swapChain: out _SwapChain
            );
        }

        private void SetUpRasterizer(int width, int height)
        {
            Viewport = new SlimDX.Direct3D10.Viewport(0, 0, width, height, 0.0f, 1.0f);
            Device.Rasterizer.SetViewports(Viewport);
            Device.Rasterizer.State = SlimDX.Direct3D10.RasterizerState.FromDescription(Device, new SlimDX.Direct3D10.RasterizerStateDescription()
                {
                    CullMode = SlimDX.Direct3D10.CullMode.Back,
                    FillMode = SlimDX.Direct3D10.FillMode.Solid,
                    IsFrontCounterclockwise = true
                });
        }

        private void SetUpOutputMerger()
        {
            Device.OutputMerger.DepthStencilState = depthStencilState;
            Device.OutputMerger.SetTargets(DepthStencilView, RenderTargetView);
        }

        SlimDX.Direct3D10.DepthStencilState depthStencilState;

        private void SetUpDepthStencilBuffer(int width, int height)
        {
            var depthBufferDescription = new SlimDX.Direct3D10.Texture2DDescription()
            {
                Width = width,
                Height = height,
                ArraySize = 1,
                MipLevels = 1,
                BindFlags = SlimDX.Direct3D10.BindFlags.DepthStencil,
                Format = SlimDX.DXGI.Format.D32_Float,
                CpuAccessFlags = SlimDX.Direct3D10.CpuAccessFlags.None,
                OptionFlags = SlimDX.Direct3D10.ResourceOptionFlags.None,
                SampleDescription = new SlimDX.DXGI.SampleDescription(1, 0),
                Usage = SlimDX.Direct3D10.ResourceUsage.Default
            };

            var dssd = new SlimDX.Direct3D10.DepthStencilStateDescription
            {
                IsDepthEnabled = true,
                IsStencilEnabled = false,
                DepthWriteMask = SlimDX.Direct3D10.DepthWriteMask.All,
                DepthComparison = SlimDX.Direct3D10.Comparison.Less
            };


             depthStencilState = SlimDX.Direct3D10.DepthStencilState.FromDescription(Device, dssd);
            


            using (var depthBuffer = new SlimDX.Direct3D10.Texture2D(Device, depthBufferDescription))
            {
                DepthStencilView = new SlimDX.Direct3D10.DepthStencilView(Device, depthBuffer);
            }
        }

        public void Clear(Color color)
        {
            Device.ClearDepthStencilView(DepthStencilView, SlimDX.Direct3D10.DepthStencilClearFlags.Depth, 1.0f, 0);
            Device.ClearRenderTargetView(RenderTargetView, color.ToSystemDrawingColor());
        }

        public void Present()
        {
            SwapChain.Present(0, SlimDX.DXGI.PresentFlags.None);
        }

        public void DrawUserPrimitives<T>(PrimitiveType primitiveType, T[] data, int offset, int primitiveCount, IAurumEffect effect )
        {
            int primitiveCountFactor = 0;

            SlimDX.Direct3D10.PrimitiveTopology topology;

            if (primitiveType == PrimitiveType.LineList)
            {
                primitiveCountFactor = 2;
                topology = SlimDX.Direct3D10.PrimitiveTopology.LineList;
            }
            else
            {
                throw new NotImplementedException();
            }

            if (typeof(T) == typeof(VertexPositionColor))
            {
                int vertexCount = primitiveCount * primitiveCountFactor;

                using (var vb = CreateVertexBuffer(VertexPositionColor.VertexDeclaration, vertexCount))
                {

                    var size = Marshal.SizeOf(typeof(T)) * vertexCount;
                    byte[] bytes = new byte [size];

                    unsafe
                    {
                        fixed (void* pointer = &((data as VertexPositionColor[])[0]))
                        {
                            Marshal.Copy(new IntPtr(pointer), bytes, 0, size);
                        }
                    }

                    vb.SetData(bytes);
                    vb.Effect = effect;
                    vb.Create();

                    vb.Apply();

                    bytes = null;

                    Device.InputAssembler.SetPrimitiveTopology(topology);
                    Device.Draw(vertexCount, 0);
                }
            }
        }

        public byte[] RawSerializer<T>(T[] data)
        {
            using (var memStream = new MemoryStream())
            {
                var values = from item in data
                             from property in typeof(T).GetProperties()
                             select property.GetValue(item, null);


                return memStream.ToArray();
            }
        }

        #endregion

        #region IDisposable Members

        protected virtual void Dispose(bool cleanManaged)
        {
            if (cleanManaged)
            {

            }

            if (SwapChain != null)
            {
                SwapChain.Dispose();
                _SwapChain = null;
            }

            if (RenderTargetView != null)
            {
                RenderTargetView.Dispose();
                RenderTargetView = null;
            }

            if (DepthStencilView != null)
            {
                DepthStencilView.Dispose();
                DepthStencilView = null;
            }

            if (_Device != null)
            {
                _Device.Dispose();
                _Device = null;
            }
        }

        public void Dispose()
        {
            Dispose(true);
        }

        #endregion
    }
}
