﻿using System.Linq;
using MachinaAurum.Xna.Graphics;
using Microsoft.Xna.Framework.Graphics;

namespace MachinaAurum.Xna.DirectX10
{
    public class DirectX10VertexBuffer : AurumVertexBuffer
    {
        private DirectX10GraphicDevice Device;
        private Microsoft.Xna.Framework.Graphics.VertexDeclaration VertexDeclaration;
        private int VertexCount;
        private byte[] Data;

        public int SizeInBytes { get { return VertexDeclaration.VertexStride * VertexCount; } }
        public SlimDX.Direct3D10.Buffer Buffer { get; private set; }
        public SlimDX.Direct3D10.InputLayout InputLayouts { get; private set; }

        public DirectX10VertexBuffer(DirectX10GraphicDevice device, VertexDeclaration vertexDeclaration, int vertexCount)
        {
            Device = device;
            this.VertexDeclaration = vertexDeclaration;
            this.VertexCount = vertexCount;
        }

        public override void SetData(byte[] data)
        {
            Data = data;
        }

        public override void Create()
        {
            var inputs = VertexDeclaration.GetVertexElements().Select((x, i) => Convert(i, x)).ToArray();
            InputLayouts = new SlimDX.Direct3D10.InputLayout(Device.Device, Effect.Signature as SlimDX.D3DCompiler.ShaderSignature, inputs);

            using (var stream = new SlimDX.DataStream(SizeInBytes, true, true))
            {
                stream.WriteRange<byte>(Data, 0, SizeInBytes);
                stream.Position = 0;

                Buffer = new SlimDX.Direct3D10.Buffer(Device.Device, stream, new SlimDX.Direct3D10.BufferDescription()
                {
                    BindFlags = SlimDX.Direct3D10.BindFlags.VertexBuffer,
                    CpuAccessFlags = SlimDX.Direct3D10.CpuAccessFlags.None,
                    OptionFlags = SlimDX.Direct3D10.ResourceOptionFlags.None,
                    SizeInBytes = SizeInBytes,
                    Usage = SlimDX.Direct3D10.ResourceUsage.Default
                });

                Data = null;
            }
        }

        private SlimDX.Direct3D10.InputElement Convert(int i, VertexElement x)
        {
            SlimDX.DXGI.Format format;

            switch (x.VertexElementFormat)
            {
                case VertexElementFormat.Byte4:
                    format = SlimDX.DXGI.Format.Unknown;
                    break;
                case VertexElementFormat.Color:
                    format = SlimDX.DXGI.Format.R8G8B8A8_UNorm;
                    break;
                case VertexElementFormat.HalfVector2:
                    format = SlimDX.DXGI.Format.Unknown;
                    break;
                case VertexElementFormat.HalfVector4:
                    format = SlimDX.DXGI.Format.Unknown;
                    break;
                case VertexElementFormat.NormalizedShort2:
                    format = SlimDX.DXGI.Format.Unknown;
                    break;
                case VertexElementFormat.NormalizedShort4:
                    format = SlimDX.DXGI.Format.Unknown;
                    break;
                case VertexElementFormat.Short2:
                    format = SlimDX.DXGI.Format.R16_SInt;
                    break;
                case VertexElementFormat.Short4:
                    format = SlimDX.DXGI.Format.R32_SInt;
                    break;
                case VertexElementFormat.Single:
                    format = SlimDX.DXGI.Format.R16_Float;
                    break;
                case VertexElementFormat.Vector2:
                    format = SlimDX.DXGI.Format.R32G32_Float;
                    break;
                case VertexElementFormat.Vector3:
                    format = SlimDX.DXGI.Format.R32G32B32_Float;
                    break;
                case VertexElementFormat.Vector4:
                    format = SlimDX.DXGI.Format.R32G32B32A32_Float;
                    break;
                default:
                    format = SlimDX.DXGI.Format.Unknown;
                    break;
            }

            return new SlimDX.Direct3D10.InputElement(x.VertexElementUsage.ToString(), 0, format, x.Offset, 0);
        }

        public override void Apply()
        {
            Device.Device.InputAssembler.SetInputLayout(InputLayouts);
            Device.Device.InputAssembler.SetPrimitiveTopology(SlimDX.Direct3D10.PrimitiveTopology.TriangleList);
            Device.Device.InputAssembler.SetVertexBuffers(0, new SlimDX.Direct3D10.VertexBufferBinding(Buffer, VertexDeclaration.VertexStride, 0));
        }

        protected override void Dispose(bool cleanManaged)
        {
            if (cleanManaged)
            {
                Data = null;

                VertexDeclaration.Dispose();
                VertexDeclaration = null;
            }

            if ( Buffer != null )
            {
                Buffer.Dispose();
                Buffer = null;
            }

            if (InputLayouts != null)
            {
                InputLayouts.Dispose();
                InputLayouts = null;
            }
        }
    }
}
