﻿using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Linq;
using System.Reflection;
using MachinaAurum.Xna.Graphics.ContentTypeReaders;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

namespace MachinaAurum.Xna.Graphics
{
    public class AurumGraphicsDeviceManager : IGraphicsDeviceService, IGraphicsDeviceManager, IDisposable
    {
        [ImportMany]
        IEnumerable<Lazy<IGraphicDevice, IGraphicDeviceMetadata>> GraphicDevices;

        Game Game;

        public AurumGraphicsDeviceManager(Game game)
        {
            Game = game;
            GraphicDevices = Enumerable.Empty<Lazy<IGraphicDevice, IGraphicDeviceMetadata>>();
        }

        public void Start(GraphicsDeviceApi deviceApi)
        {
            var device = GraphicDevices.Where(x => x.Metadata.Api == deviceApi.ToString()).SingleOrDefault();

            if (device == null)
            {
                throw new DeviceNotFoundException();
            }

            SetUp(device);
        }

        private void SetUp(Lazy<IGraphicDevice, IGraphicDeviceMetadata> device)
        {
            Game.Services.AddService(typeof(IGraphicsDeviceManager), this);
            Game.Services.AddService(typeof(IGraphicsDeviceService), this);

            MachinaAurum.Xna.Devices.Graphic = device.Value;
            MachinaAurum.Xna.Devices.Graphic.Create(Game);

            var deviceCreated = DeviceCreated;
            if (deviceCreated != null)
            {
                deviceCreated(this, EventArgs.Empty);
            }

            FakeContentReaders();
        }

       
        private void FakeContentReaders()
        {
            var vertexBufferReaderType = typeof(BasicEffect).Assembly.GetType("Microsoft.Xna.Framework.Content.VertexBufferReader");
            var indexBufferReaderType = typeof(BasicEffect).Assembly.GetType("Microsoft.Xna.Framework.Content.IndexBufferReader");
            var effectReaderType = typeof(BasicEffect).Assembly.GetType("Microsoft.Xna.Framework.Content.BasicEffectReader");
            var modelReaderType = typeof(BasicEffect).Assembly.GetType("Microsoft.Xna.Framework.Content.ModelReader");

            var vbReaderInstance = new AurumVertexBufferReader(Devices.VertexBufferReader);
            var ibReaderInstance = Devices.IndexBufferReader as ContentTypeReader;
            var effReaderInstance = Devices.EffectReader as ContentTypeReader;
            var modelReaderInstance = Devices.ModelReader as ContentTypeReader;

            // FAKE VERTEX BUFFER 

            {
                var nameToReader = typeof(ContentTypeReaderManager).GetField("nameToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<string, ContentTypeReader>;
                nameToReader.Add("Microsoft.Xna.Framework.Content.VertexBufferReader, Microsoft.Xna.Framework.Graphics, Version=4.0.0.0, Culture=neutral, PublicKeyToken=6d5c3888ef60e27d", vbReaderInstance);

                var readerTypeToReader = typeof(ContentTypeReaderManager).GetField("readerTypeToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<Type, ContentTypeReader>;
                readerTypeToReader.Add(vertexBufferReaderType, vbReaderInstance);

                var targetTypeToReader = typeof(ContentTypeReaderManager).GetField("targetTypeToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<Type, ContentTypeReader>;
                targetTypeToReader.Add(vertexBufferReaderType, vbReaderInstance);
            }

            // FAKE INDEX BUFFER

            {
                var nameToReader = typeof(ContentTypeReaderManager).GetField("nameToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<string, ContentTypeReader>;
                nameToReader.Add("Microsoft.Xna.Framework.Content.IndexBufferReader, Microsoft.Xna.Framework.Graphics, Version=4.0.0.0, Culture=neutral, PublicKeyToken=6d5c3888ef60e27d", ibReaderInstance);

                var readerTypeToReader = typeof(ContentTypeReaderManager).GetField("readerTypeToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<Type, ContentTypeReader>;
                readerTypeToReader.Add(indexBufferReaderType, ibReaderInstance);

                var targetTypeToReader = typeof(ContentTypeReaderManager).GetField("targetTypeToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<Type, ContentTypeReader>;
                targetTypeToReader.Add(indexBufferReaderType, ibReaderInstance);
            }

            // FAKE BASIC EFFECT

            {
                var nameToReader = typeof(ContentTypeReaderManager).GetField("nameToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<string, ContentTypeReader>;
                nameToReader.Add("Microsoft.Xna.Framework.Content.BasicEffectReader, Microsoft.Xna.Framework.Content, Version=4.0.0.0, Culture=neutral, PublicKeyToken=6d5c3888ef60e27d", effReaderInstance);

                var readerTypeToReader = typeof(ContentTypeReaderManager).GetField("readerTypeToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<Type, ContentTypeReader>;
                readerTypeToReader.Add(effectReaderType, effReaderInstance);

                var targetTypeToReader = typeof(ContentTypeReaderManager).GetField("targetTypeToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<Type, ContentTypeReader>;
                targetTypeToReader.Add(effectReaderType, effReaderInstance);
            }

            // FAKE MODEL READER

            {
                var nameToReader = typeof(ContentTypeReaderManager).GetField("nameToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<string, ContentTypeReader>;
                nameToReader.Add("Microsoft.Xna.Framework.Content.ModelReader, Microsoft.Xna.Framework.Content, Version=4.0.0.0, Culture=neutral, PublicKeyToken=6d5c3888ef60e27d", modelReaderInstance);

                var readerTypeToReader = typeof(ContentTypeReaderManager).GetField("readerTypeToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<Type, ContentTypeReader>;
                readerTypeToReader.Add(modelReaderType, modelReaderInstance);

                var targetTypeToReader = typeof(ContentTypeReaderManager).GetField("targetTypeToReader", BindingFlags.Static | BindingFlags.NonPublic).GetValue(null) as Dictionary<Type, ContentTypeReader>;
                targetTypeToReader.Add(modelReaderType, modelReaderInstance);
            }
        }

        #region IGraphicsDeviceService Members

        public event EventHandler<EventArgs> DeviceCreated;
        public event EventHandler<EventArgs> DeviceDisposing;
        public event EventHandler<EventArgs> DeviceReset;
        public event EventHandler<EventArgs> DeviceResetting;

        public GraphicsDevice GraphicsDevice
        {
            get { return null; }
        }

        #endregion

        #region IGraphicsDeviceManager Members

        public bool BeginDraw()
        {
            return true;
        }

        public void CreateDevice()
        {

        }

        public void EndDraw()
        {
            MachinaAurum.Xna.Devices.Graphic.Present();
        }

        #endregion

        #region IDisposable Members

        protected virtual void Dispose(bool cleanManaged)
        {
            if (cleanManaged)
            {

            }

            if (MachinaAurum.Xna.Devices.Graphic != null)
            {
                MachinaAurum.Xna.Devices.Graphic.Dispose();
                MachinaAurum.Xna.Devices.Graphic = null;
            }
        }

        public void Dispose()
        {
            Dispose(true);
        }

        #endregion
    }
}
