﻿using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

namespace MachinaAurum.Xna.Graphics
{
    public class AurumModel : IAurumModel
    {
        private List<IAurumModelBone> bones;
        private List<IAurumModelMesh> meshes;
        private static Matrix[] sharedDrawBoneMatrices;

        public object Tag { get; set; }
        public IAurumModelBone Root { get; private set; }
        public IEnumerable<IAurumModelBone> Bones
        {
            get
            {
                return this.bones;
            }
        }
        public IEnumerable<IAurumModelMesh> Meshes
        {
            get
            {
                return this.meshes;
            }
        }


        public AurumModel()
        {
        }

        public void CopyAbsoluteBoneTransformsTo(Microsoft.Xna.Framework.Matrix[] destinationBoneTransforms)
        {
            if (destinationBoneTransforms == null)
            {
                throw new ArgumentNullException("destinationBoneTransforms");
            }
            if (destinationBoneTransforms.Length < this.bones.Count)
            {
                throw new ArgumentOutOfRangeException("destinationBoneTransforms");
            }
            int count = this.bones.Count;
            for (int i = 0; i < count; i++)
            {
                IAurumModelBone bone = this.bones[i];
                if (bone.Parent == null)
                {
                    destinationBoneTransforms[i] = bone.Transform;
                }
                else
                {
                    int index = bone.Parent.Index;
                    destinationBoneTransforms[i] = bone.Transform * destinationBoneTransforms[index];
                }
            }
        }

        public void CopyBoneTransformsFrom(Microsoft.Xna.Framework.Matrix[] sourceBoneTransforms)
        {
            if (sourceBoneTransforms == null)
            {
                throw new ArgumentNullException("sourceBoneTransforms");
            }
            if (sourceBoneTransforms.Length < this.bones.Count)
            {
                throw new ArgumentOutOfRangeException("sourceBoneTransforms");
            }
            int count = this.bones.Count;
            for (int i = 0; i < count; i++)
            {
                this.bones[i].Transform = sourceBoneTransforms[i];
            }
        }

        public void CopyBoneTransformsTo(Microsoft.Xna.Framework.Matrix[] destinationBoneTransforms)
        {
            if (destinationBoneTransforms == null)
            {
                throw new ArgumentNullException("destinationBoneTransforms");
            }
            if (destinationBoneTransforms.Length < this.bones.Count)
            {
                throw new ArgumentOutOfRangeException("destinationBoneTransforms");
            }
            int count = this.bones.Count;
            for (int i = 0; i < count; i++)
            {
                destinationBoneTransforms[i] = this.bones[i].Transform;
            }
        }

        public void Draw(Microsoft.Xna.Framework.Matrix world, Microsoft.Xna.Framework.Matrix view, Microsoft.Xna.Framework.Matrix projection)
        {
            int count = this.meshes.Count;
            int num3 = this.bones.Count;
            Microsoft.Xna.Framework.Matrix[] sharedDrawBoneMatrices = AurumModel.sharedDrawBoneMatrices;
            if ((sharedDrawBoneMatrices == null) || (sharedDrawBoneMatrices.Length < num3))
            {
                sharedDrawBoneMatrices = new Microsoft.Xna.Framework.Matrix[num3];
                AurumModel.sharedDrawBoneMatrices = sharedDrawBoneMatrices;
            }
            this.CopyAbsoluteBoneTransformsTo(sharedDrawBoneMatrices);
            for (int i = 0; i < count; i++)
            {
                IAurumModelMesh mesh = this.meshes[i];
                int index = mesh.ParentBone.Index;

                foreach (var effect in mesh.Effects)
                {
                    if (effect == null)
                    {
                        throw new InvalidOperationException("ModelHasNoEffect");
                    }
                    IEffectMatrices matrices = effect as IEffectMatrices;
                    if (matrices == null)
                    {
                        throw new InvalidOperationException("ModelHasNoIEffectMatrices");
                    }
                    matrices.World = sharedDrawBoneMatrices[index] * world;
                    matrices.View = view;
                    matrices.Projection = projection;
                }

                mesh.Draw();
            }
        }

        public static AurumModel Read(ContentReader input)
        {
            AurumModel model = new AurumModel();

            model.ReadBones(input);
            model.ReadMeshes(input);
            model.Root = model.ReadBoneReference(input);
            model.Tag = input.ReadObject<object>();

            return model;
        }

        private IAurumModelBone ReadBoneReference(ContentReader input)
        {
            int num2 = this.bones.Count + 1;
            int num = 0;
            if (num2 <= 0xff)
            {
                num = input.ReadByte();
            }
            else
            {
                num = input.ReadInt32();
            }
            if (num != 0)
            {
                return this.bones[num - 1];
            }
            return null;
        }

        private void ReadBones(ContentReader input)
        {
            AurumModelBone[] bones = new AurumModelBone[input.ReadInt32()];
            for (int i = 0; i < bones.Length; i++)
            {
                string name = input.ReadObject<string>();
                Microsoft.Xna.Framework.Matrix transform = input.ReadMatrix();
                bones[i] = new AurumModelBone(name, transform, i);
            }
            this.bones = new List<IAurumModelBone>(bones);
            foreach (AurumModelBone bone2 in bones)
            {
                IAurumModelBone newParent = this.ReadBoneReference(input);
                int num4 = input.ReadInt32();
                IAurumModelBone[] newChildren = new AurumModelBone[num4];
                for (int j = 0; j < num4; j++)
                {
                    newChildren[j] = this.ReadBoneReference(input);
                }
                bone2.SetParentAndChildren(newParent, newChildren);
            }
        }

        private void ReadMeshes(ContentReader input)
        {
            int num2 = input.ReadInt32();
            AurumModelMesh[] meshes = new AurumModelMesh[num2];
            for (int i = 0; i < num2; i++)
            {
                string name = input.ReadObject<string>();
                IAurumModelBone parentBone = this.ReadBoneReference(input);
                Microsoft.Xna.Framework.BoundingSphere boundingSphere = new Microsoft.Xna.Framework.BoundingSphere
                {
                    Center = input.ReadVector3(),
                    Radius = input.ReadSingle()
                };
                object tag = input.ReadObject<object>();
                AurumModelMeshPart[] meshParts = ReadMeshParts(input);
                meshes[i] = new AurumModelMesh(name, parentBone, boundingSphere, meshParts, tag);
            }
            this.meshes = new List<IAurumModelMesh>(meshes);
        }

        private static AurumModelMeshPart[] ReadMeshParts(ContentReader input)
        {
            int num2 = input.ReadInt32();
            AurumModelMeshPart[] meshParts = new AurumModelMeshPart[num2];
            for (int i = 0; i < num2; i++)
            {
                int vertexOffset = input.ReadInt32();
                int numVertices = input.ReadInt32();
                int startIndex = input.ReadInt32();
                int primitiveCount = input.ReadInt32();
                object tag = input.ReadObject<object>();
                meshParts[i] = new AurumModelMeshPart(vertexOffset, numVertices, startIndex, primitiveCount, tag);
                int uniqueCopyOfI = i;
                input.ReadSharedResource<AurumVertexBuffer>(delegate(AurumVertexBuffer vb)
                {
                    meshParts[uniqueCopyOfI].VertexBuffer = vb;
                });
                input.ReadSharedResource<AurumIndexBuffer>(delegate(AurumIndexBuffer ib)
                {
                    meshParts[uniqueCopyOfI].IndexBuffer = ib;
                });
                input.ReadSharedResource<AurumEffect>(delegate(AurumEffect effect)
                {
                    meshParts[uniqueCopyOfI].Effect = effect;
                    meshParts[uniqueCopyOfI].Effect.Create();

                    meshParts[uniqueCopyOfI].VertexBuffer.Effect = effect;
                    meshParts[uniqueCopyOfI].VertexBuffer.Create();

                    meshParts[uniqueCopyOfI].IndexBuffer.Create();
                });
            }
            return meshParts;
        }

    }
}
