﻿using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Web.Razor;
using Microsoft.CSharp;

namespace MachinaAurum.TemplateEngine
{
    class RazorTemplateEngine : ITemplateEngine
    {
        System.Web.Razor.RazorTemplateEngine Engine;
        IDictionary<string, Func<dynamic>> InstanceFactory;

        public RazorTemplateEngine()
        {
            Engine = SetupRazorEngine();
            InstanceFactory = new Dictionary<string, Func<dynamic>>();
        }

        public dynamic GetTemplate(string name)
        {
            if (InstanceFactory.ContainsKey(name) == false)
            {
                Create(name);
            }

            return InstanceFactory[name]();
        }

        private System.Web.Razor.RazorTemplateEngine SetupRazorEngine()
        {
            RazorEngineHost host = new RazorEngineHost(new CSharpRazorCodeLanguage());
            host.DefaultBaseClass = typeof(TemplateBase).FullName;
            host.DefaultNamespace = "RazorOutput";
            host.DefaultClassName = "Template";
            host.NamespaceImports.Add("System");
            return new System.Web.Razor.RazorTemplateEngine(host);
        }

        private void Create(string name)
        {
            string template = Assembly.GetExecutingAssembly().GetEmbeddedResourceByName(name).EncodeToString();

            GeneratorResults razorResult = null;
            using (TextReader rdr = new StringReader(template))
            {
                razorResult = Engine.GenerateCode(rdr);
            }

            CSharpCodeProvider codeProvider = new CSharpCodeProvider();

            //Generate the code
            string debugCode = string.Empty;
            using (StringWriter sw = new StringWriter())
            {
                codeProvider.GenerateCodeFromCompileUnit(razorResult.GeneratedCode, sw, new CodeGeneratorOptions());
                debugCode = sw.GetStringBuilder().ToString();
            }


            CompilerResults results = codeProvider.CompileAssemblyFromDom(new CompilerParameters
            (
                new string[]
                {
                    Assembly.GetExecutingAssembly().CodeBase.Replace("file:///", "").Replace("/", "\\")
                }
            ), razorResult.GeneratedCode);

            if (results.Errors.HasErrors)
            {
                CompilerError err = results.Errors
                                           .OfType<CompilerError>()
                                           .Where(ce => !ce.IsWarning)
                                           .First();

                throw new Exception();
            }
            else
            {
                Assembly asm = results.CompiledAssembly;
                Type type = asm.GetType("RazorOutput.Template");
                InstanceFactory.Add(name, () =>
                {
                    var instance = Activator.CreateInstance(type) as dynamic;
                    instance.Engine = this;

                    return instance;
                });
            }
        }

    }


    public abstract class TemplateBase
    {
        public StringBuilder Buffer { get; set; }
        public StringWriter Writer { get; set; }

        public ITemplateEngine Engine { get; set; }

        public TemplateBase()
        {
            Buffer = new StringBuilder();
            Writer = new StringWriter(Buffer);
        }

        public abstract void Execute();

        // Writes the results of expressions like: "@foo.Bar"
        public virtual void Write(object value)
        {
            // Don't need to do anything special
            // Razor for ASP.Net does HTML encoding here.
            WriteLiteral(value);
        }

        // Writes literals like markup: "<p>Foo</p>"
        public virtual void WriteLiteral(object value)
        {
            Buffer.Append(value);
        }

        protected string PartialView(string templateName, object viewModel)
        {
            dynamic template = Engine.GetTemplate(templateName);
            try
            {
                template.ViewModel = viewModel;
            }
            catch
            { }
        
            template.Execute();
            return template.ToString();
        }

        public override string ToString()
        {
            return Buffer.ToString();
        }
    }
}
