﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text.RegularExpressions;
using Microsoft.Practices.Unity;

namespace MachinaAurum.Unity.Aspect
{
    public class UnityAopExtension : UnityContainerExtension
    {
        public UnityAopExtension RegisterType<TFrom, TTo>()
        {
            Type fromType = typeof(TFrom);
            Type to = typeof(TTo);
            Type fromProxy = fromType.Assembly.GetType("Proxy" + fromType.Name) ?? to.Assembly.GetType("Proxy" + to.Name);

            if (fromProxy == null)
            {
                throw new Exception("Proxy not found");
            }

            Container.RegisterType<TFrom>(new InjectionFactory((u) =>
                {
                    var proxy = u.Resolve(fromProxy, new ParameterOverride("target", u.Resolve<TTo>()));

                    InjectMethods(fromType, to, u, proxy);
                    InjectProperties(fromType, to, u, proxy);

                    return proxy;
                }));

            return this;
        }

        private void InjectMethods(Type fromType, Type to, IUnityContainer u, object proxy)
        {
            var injectionAttributesInAssembly = from att in fromType.Assembly.GetCustomAttributes(false).OfType<InjectAttribute>()
                                                from method in fromType.GetMethods().Where(m => m.IsSpecialName == false).ToList()
                                                from rule in att.Rules
                                                where Regex.IsMatch(method.ReflectedType.FullName + "." + method.Name, rule)
                                                select new
                                                {
                                                    Method = method,
                                                    Attribute = att
                                                };

            var injectionAttributesInClass = from att in to.GetCustomAttributes(false).OfType<InjectAttribute>()
                                             from fromMethod in fromType.GetMethods().Where ( m => m.IsSpecialName == false ).ToList()
                                             from rule in att.Rules
                                             let method = to.GetMethod(fromMethod.Name, fromMethod.GetParameters().Select(p => p.ParameterType).ToArray())
                                             where Regex.IsMatch(method.Name, rule)
                                             select new
                                             {
                                                 Method = method,
                                                 Attribute = att
                                             };

            var injectionAttributesInToType = from method in to.GetMethods().Where(m => m.IsSpecialName == false).ToList()
                                              from att in method.GetCustomAttributes(false).OfType<InjectAttribute>()
                                              select new
                                              {
                                                  Method = method,
                                                  Attribute = att
                                              };

            var customAttributes = UnionAll(UnionAll(injectionAttributesInAssembly, injectionAttributesInClass), injectionAttributesInToType).OrderBy(a => a.Attribute.Order);

            foreach (var item in customAttributes)
            {
                var parameterOverrides = item.Attribute.Parameters.Select(k => new ParameterOverride(k.Key, k.Value)).ToArray();
                IMethodInjection injectInstance = u.Resolve(item.Attribute.Injection, parameterOverrides) as IMethodInjection;

                string methodName = item.Method.Name;
                InjectBeforeMethod(proxy, injectInstance, methodName);
                InjectAfterMethod(proxy, injectInstance, methodName);
            }
        }

        private void InjectProperties(Type fromType, Type to, IUnityContainer u, object proxy)
        {
            var injectionAttributesInAssembly = from att in fromType.Assembly.GetCustomAttributes(false).OfType<InjectAttribute>()
                                                from property in fromType.GetProperties().ToList()
                                                from rule in att.Rules
                                                where Regex.IsMatch(property.ReflectedType.FullName + "." + property.Name, rule)
                                                select new
                                                {
                                                    Property = property,
                                                    Attribute = att
                                                };

            var injectionAttributesInClass = from att in to.GetCustomAttributes(false).OfType<InjectAttribute>()
                                             from fromProperty in fromType.GetProperties().ToList()
                                             from rule in att.Rules
                                             let property = to.GetProperty(fromProperty.Name)
                                             where Regex.IsMatch(property.Name, rule)
                                             select new
                                             {
                                                 Property = property,
                                                 Attribute = att
                                             };

            var injectionAttributesInToType = from property in to.GetProperties().ToList()
                                              from att in property.GetCustomAttributes(false).OfType<InjectAttribute>()
                                              select new
                                              {
                                                  Property = property,
                                                  Attribute = att
                                              };

            var customAttributes = UnionAll(UnionAll(injectionAttributesInAssembly, injectionAttributesInClass), injectionAttributesInToType).OrderBy(a => a.Attribute.Order);

            foreach (var item in customAttributes)
            {
                var parameterOverrides = item.Attribute.Parameters.Select(k => new ParameterOverride(k.Key, k.Value)).ToArray();
                IPropertyInjection injectInstance = u.Resolve(item.Attribute.Injection, parameterOverrides) as IPropertyInjection;

                string propertyName = item.Property.Name;

                InjectBeforeGet(proxy, injectInstance, propertyName);
                InjectAfterGet(proxy, injectInstance, propertyName);

                InjectBeforeSet(proxy, injectInstance, propertyName);
                InjectAfterSet(proxy, injectInstance, propertyName);
            }
        }

        IEnumerable<T> UnionAll<T>(IEnumerable<T> first, IEnumerable<T> second)
        {
            foreach (var item in first)
            {
                yield return item;
            }

            foreach (var item in second)
            {
                yield return item;
            }
        }

        private static void InjectBeforeMethod(object proxy, IMethodInjection injectInstance, string methodName)
        {
            string beforeFunctionName = "Before" + methodName;
            proxy.GetType().InvokeMember(beforeFunctionName, BindingFlags.InvokeMethod | BindingFlags.Public | BindingFlags.Instance, null, proxy, new object[]
                        {
                            new Action<MethodInfo,object[]>((m, p) =>{injectInstance.Before(m, p);})
                        });
        }

        private static void InjectAfterMethod(object proxy, IMethodInjection injectInstance, string methodName)
        {
            string afterFunctionName = "After" + methodName;
            proxy.GetType().InvokeMember(afterFunctionName, BindingFlags.InvokeMethod | BindingFlags.Public | BindingFlags.Instance, null, proxy, new object[]
                        {
                            new Action<MethodInfo,object[]>((m, p) =>{injectInstance.After(m, p);})
                        });
        }

        private static void InjectBeforeGet(object proxy, IPropertyInjection injectInstance, string propertyName)
        {
            string beforeFunctionName = "BeforeGet" + propertyName;
            proxy.GetType().InvokeMember(beforeFunctionName, BindingFlags.InvokeMethod | BindingFlags.Public | BindingFlags.Instance, null, proxy, new object[]
                        {
                            new Action<PropertyInfo>( p =>{injectInstance.BeforeGet(p);} )
                        });
        }

        private static void InjectAfterGet(object proxy, IPropertyInjection injectInstance, string propertyName)
        {
            string beforeFunctionName = "AfterGet" + propertyName;
            proxy.GetType().InvokeMember(beforeFunctionName, BindingFlags.InvokeMethod | BindingFlags.Public | BindingFlags.Instance, null, proxy, new object[]
                        {
                            new Action<PropertyInfo>( p =>{injectInstance.AfterGet(p);} )
                        });
        }

        private static void InjectBeforeSet(object proxy, IPropertyInjection injectInstance, string propertyName)
        {
            string beforeFunctionName = "BeforeSet" + propertyName;
            proxy.GetType().InvokeMember(beforeFunctionName, BindingFlags.InvokeMethod | BindingFlags.Public | BindingFlags.Instance, null, proxy, new object[]
                        {
                            new Action<PropertyInfo>( p =>{injectInstance.BeforeSet(p);} )
                        });
        }

        private static void InjectAfterSet(object proxy, IPropertyInjection injectInstance, string propertyName)
        {
            string beforeFunctionName = "AfterSet" + propertyName;
            proxy.GetType().InvokeMember(beforeFunctionName, BindingFlags.InvokeMethod | BindingFlags.Public | BindingFlags.Instance, null, proxy, new object[]
                        {
                            new Action<PropertyInfo>( p =>{injectInstance.AfterSet(p);} )
                        });
        }


        protected override void Initialize()
        {
        }
    }
}
